/// <reference types="node" />
import * as xlib from "xlib";
import * as _requestLib from "request";
import __ = xlib.lolo;
export import URL = require("url");
import * as __request from "request";
export { __request };
export declare type IRequestOptions = {
    /** if not explicitly set,  we set ```requestLibOptions.encoding=null```.  So the response body is a Buffer.  pass ```true``` here to use the requestLib default, which is to convert to utf8 */
    _slibDisableDefaultEncoding?: boolean;
} & __request.Options;
/** if a technical failure is encountered when processing the request, this is the rejection type.
 * Note that a response statusCode of 400+ is not a technical failure (it will be be returned normally)
    */
export declare class RequestException extends __.Ex {
    response: __request.Response;
    requestOptions: IRequestOptions;
    /** the classification of the technical failure.   check  [[message]] for more details.
     *
     *
     *
     * ```ECONNRESET```: could not connect properly.   maybe port is not open.
     *
     * ```ETIMEDOUT```: session timeout
     *
     * ```ESOCKETTIMEDOUT```: timeout
     *
     * ```Invalid protocol```: rejection before submitting your request.  unknown protocol specified (check url)
     *
     * ```ENOTFOUND```: dns lookup failed, etc
     *
     * ```EPROTO```: ssl handshake failure, invalid cert, etc
     *
     * ```UNKNOWN```: some other technical failure, but a code was not generated.   check [[message]] for details
     *
     *
     * ***IMPORTANT NOTE***: codes other than those specified may be returned (for various rare system level technical errors).
     * such as this list from https://www.npmjs.com/package/core-error-predicates?activeTab=readme#networkerror:
     *
    EADDRINFO
    EADDRNOTAVAIL
    EAFNOSUPPORT
    EALREADY
    ECONNABORTED
    ECONNREFUSED
    ECONNRESET
    EDESTADDRREQ
    EHOSTUNREACH
    EISCONN
    EMSGSIZE
    ENETDOWN
    ENETUNREACH
    ENONET
    ENOTCONN
    ENOTSOCK
    ENOTSUP
    EPIPE
    EPROTO
    EPROTONOSUPPORT
    EPROTOTYPE
    ETIMEDOUT
    EAIFAMNOSUPPORT
    EAISERVICE
    EAISOCKTYPE
    ESHUTDOWN
        ENOTFOUND
    */
    code: "ETIMEDOUT" | "ECONNRESET" | "ESOCKETTIMEDOUT" | "Invalid protocol" | "ENOTFOUND" | "EPROTO" | "UNKNOWN" | "ECONNABORTED" | "ECONNREFUSED" | "ENONET";
    /** Set to `true` if the timeout was a connection timeout, `false`  otherwise. */
    connect: boolean;
    constructor(message: string, _innerErr: any, response: __request.Response, requestOptions: IRequestOptions);
}
export interface IResponsePayload {
    response: __request.Response;
    body: Buffer;
}
/** make a network request.    Pass ```options.simple=false``` to get a normal response for non 2xx statusCodes.
 *
 * we wrap any errors returned from underlying calls inside a [[RequestException]] object and reject with that.
 *
 * internally we use the ```request``` library.   https://www.npmjs.com/package/request.
 *
 * We switched from ```axios``` because it has persistant bugs around proxy and httpsAgent support.
 *
 */
export declare function request(
/** ***IMPORTANT***. by default we set ```simple=false```, which means 4xx errors will not throw exceptions  */
options: IRequestOptions): Promise<IResponsePayload>;
/** invokes [[request]]() but also ensures response body is of type ```application/json```, automatically parses it, and returns it as an extra parameter.
*
* additionally, if you submit an ```options.body``` as an object, we automatically convert it to a string for submission
*/
export declare function requestJson<TResponseBody>(options: IRequestOptions & {
    /**optional.  pass TRUE to require the "content-type" header to be set to "application/json".  Default is FALSE.  */
    requireJsonHeader?: boolean;
}): Promise<{
    response: _requestLib.Response;
    body: TResponseBody;
}>;
/**
 * Sames are Node's url.parse() just adds the 'username', 'password' and 'scheme' fields.
 *
 * Note that `scheme` is always lower-cased (e.g. `ftp`).
 * @param url
 * @ignore
 */
export declare function parseUrl(url: string): URL.UrlWithStringQuery & {
    username?: string;
    password?: string;
    scheme?: string;
};
/** vm specific network features.   These work on google cloud, but some work on other providers */
export declare namespace vm {
    /** External IP address of the VM.   Works on google cloud, AWS, or Azure.
     *
     * solution from: https://stackoverflow.com/questions/23362887/can-you-get-external-ip-address-from-within-a-google-compute-vm-instance
    
    for example, on google you could do ```curl --header "Metadata-Flavor:Google" http://metadata/computeMetadata/v1/instance/network-interfaces/0/access-configs/0/external-ip```
    */
    function getExternalIp(): Promise<string>;
    /**
 *  extract the final client ip address, for use either when using a "bare" server or when using Google Compute's HTTP Load Balancer.
 *
 * @param req
 */
    function extractClientIpBareOrGCloudHttpLoadBalancer(
    /** works directly with hapi's request object */
    req: {
        info: {
            remoteAddress: string;
        };
        headers: Record<string, string>;
    }, forwardingSteps?: number, proxyIfRemoteAddressStartsWith?: string, ipIfInvalid?: string): string;
}
//# sourceMappingURL=net.d.ts.map