"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const xlib = tslib_1.__importStar(require("xlib"));
var __ = xlib.lolo;
const readline = tslib_1.__importStar(require("readline"));
/** cross platform shell commands.  not promise based unfortunately  */
const shelljs = require("shelljs");
const chalk = xlib.util.stringHelper.Chalk.default;
/** ask a question on the console, and awaits a typed answer (followed by the ```ENTER``` key) */
async function question(query, options = {}) {
    if (query.trim().endsWith(":") !== true) {
        query = "\n" + query.trim() + ": ";
    }
    options = Object.assign({}, options);
    while (true) {
        let answer = await __question(chalk.yellow(query));
        if (options.forceLowercase === true) {
            answer = answer.toLowerCase();
        }
        if (__.str.isNullOrEmpty(answer) === false) {
            return answer;
        }
        else {
            //blank answer
            if (options.defaultIfBlank != null) {
                //default answer
                answer = options.defaultIfBlank;
                return answer;
            }
            else {
                comment("blank answers are not allowed for this question.");
                continue;
            }
        }
    }
}
exports.question = question;
/** bare-bones question prompt. as this isn't very user friendly, generally you should NOT use this directly.  instead use ```askQuestion()``` or ```askQuestionChoices()``` */
async function __question(query) {
    return new __.bb(async (resolve, reject) => {
        const __readStdin = readline.createInterface({ input: process.stdin, output: process.stdout });
        __readStdin.question(query, (answer) => {
            __readStdin.close();
            resolve(answer);
        });
    });
}
exports.__question = __question;
/** ask a question, offering explicit choices the user can select from.   default choices are [y,n] with no blank options allowed. */
async function questionChoices(query, options) {
    options = Object.assign({ choices: ["y", "n"] }, options);
    const baseOptions = Object.assign({}, options, { forceLowercase: options.preserveCase !== true, defaultIfBlank: options.defaultIfBlank == null ? null : options.defaultIfBlank.toString() });
    function showNumericChoices() {
        let toReturn = "";
        for (let i = 0; i < options.choices.length; i++) {
            toReturn += `${chalk.bgYellow.black((i + 1).toString())}. ${options.choices[i]}\n`;
        }
        return toReturn;
    }
    function showDefaultChoice() {
        if (options.defaultIfBlank == null) {
            return "";
        }
        return " " + chalk.bgGreen.black(options.defaultIfBlank.toString());
    }
    function showChoiceSummary() {
        if (options.numericSelect === true) {
            return `[${chalk.bgYellow.black("1")}-${chalk.bgYellow.black((options.choices.length).toString())}]${showDefaultChoice()}`;
        }
        else {
            return `[${options.choices.map((val) => chalk.bgYellow.black(val)).join()}]${showDefaultChoice()}`;
        }
    }
    while (true) {
        let answer;
        if (options.numericSelect === true) {
            answer = await (question(`${query} \n${showNumericChoices()} ${showChoiceSummary()}`, baseOptions));
            //if user selected a number, pick that choice
            let asNum = __.num.parseInt(answer, 0);
            if (asNum >= 1 && asNum <= options.choices.length) {
                answer = options.choices[asNum - 1];
            }
        }
        else {
            answer = await (question(`${query} ${showChoiceSummary()}`, baseOptions));
        }
        if (options.choices.includes(answer)) {
            return answer;
        }
        else {
            comment(`invalid choice.  you must choose from ${showChoiceSummary()}  you choose "${chalk.bgRed.white(answer)}"`);
        }
    }
}
exports.questionChoices = questionChoices;
/** execute a command to the native shell */
async function exec(cmd, options) {
    options = Object.assign({ exitCode: 0 }, options);
    shelljs.echo(chalk.bold.red("shell.exec: ") + chalk.italic.gray(cmd));
    //log.info( `about to exec "${ cmd }".  It's output is:`);
    let toReturn = new __.bb((resolve, reject) => {
        shelljs.exec(cmd, options, (code, stdout, stderr) => {
            //normalize params
            code = code == null ? -1 : code;
            stdout = stdout == null ? "" : stdout.trim();
            stderr = stderr == null ? "" : stderr.trim();
            shelljs.echo(chalk.bold.red("shell.exec,exitCode: ") + chalk.italic.gray(`${code} (expected ${options.exitCode})`));
            if (options.exitCode != null) {
                if (code !== options.exitCode) {
                    const wrongErrorCodeEx = new Error(`WRONG ERRORCODE.  Expected "${options.exitCode}" but got "${code}" when running "shell.exec('${cmd}')"`);
                    reject(wrongErrorCodeEx);
                    return;
                }
            }
            let toResolve = { shellCmd: cmd, code, stdout, stderr };
            resolve(toResolve);
            return;
        });
    });
    return toReturn;
}
exports.exec = exec;
/** write a section (heading) to console.out */
function section(...args) {
    let strs = args.map((arg) => typeof (arg) === "string" ? arg : JSON.stringify(arg, undefined, "	"));
    shelljs.echo(chalk.bold.red("===== ") + chalk.italic.black.bgYellow(...strs) + chalk.bold.red(" ====="));
}
exports.section = section;
/** write a comment (minor text) to console.out */
function comment(...args) {
    let strs = args.map((arg) => typeof (arg) === "string" ? arg : JSON.stringify(arg, undefined, "	"));
    shelljs.echo(chalk.bold.red("comment: ") + chalk.italic.gray(...strs));
}
exports.comment = comment;
/** write plain text to console.out */
function echo(...args) {
    let strs = args.map((arg) => typeof (arg) === "string" ? arg : JSON.stringify(arg));
    shelljs.echo(...strs);
}
exports.echo = echo;
//# sourceMappingURL=shell.js.map